#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""c3pcontrol - remote control of the CRONUS-3P laser.

Copyright 2020-2024 Light Conversion
Contact: support@lightcon.com
"""
import json
import time
import datetime
from enum import IntEnum
import numpy as np
from ..common._http_methods import HTTP_methods
from ..common._logging import init_logger
from ..wintopas._wintopas import WinTopas
from ..laser_clients._laser_clients import Carbide


class C3PState(IntEnum):
    """CRONUS-3P laser states."""
    UnknownState = -3
    Fail = -2
    SoftFail = -1
    Idle = 0
    Preparation = 1
    MovingMotorsToInitialPosition = 2
    AdjustingOutputBeam = 3
    Success = 4


class C3PControl_Base(HTTP_methods):
    """CRONUS-3P remote control interface class."""
    silent = True
    verbose = False
    connected = False
    logger = None
    type = 'cronus-3p'
    hw_ver = None  # c3pv1 or c3pv2, set and implemented elsewhere
    has_gdd_control = True

    def __init__(
            self, ip_address, port=35120, dev_sn=None, version='v0', **kwargs):
        """Initialization."""
        self.logger = init_logger('c3p', 'cronus_3p.log')
        self.url = 'http://{}:{}/{}/{}/'.format(
            ip_address, port, dev_sn, version)
        self.logger.info("Connecting to CRONUS-3P device "
                         "SN {:s} at {:s}:{:d}".format(
                            dev_sn, ip_address, port))
        try:
            self.connected = self._open('Help/API').find('Beam') != -1
        except Exception as excpt:
            self.logger.error(
                "Could not connect to CRONUS-3P. Make sure the IP address, "
                "the REST port and the device SN are correct, the host is "
                "reachable, and that Light Composer is running. Exception "
                "reason: %s", excpt)
            self.connected = False

        if self.connected:
            self.logger.info("CRONUS-3P control established at %s", self.url)

        self.tctrl = None
        self.with_topas_control = kwargs.get('with_topas_control')
        if not self.with_topas_control:
            if kwargs.get('with_attenuator_control'):
                self.with_topas_control = True
                print("Attenuator control for this device requires a "
                      "connection to the Topas4 REST server.")
                self.idl_attn_motor_id = kwargs.get('idl_attn_motor_id')

        if self.with_topas_control:
            try:
                print("Connecting to the Topas4 server of CRONUS-3P...")
                self.tctrl = WinTopas(ip_address, dev_sn=dev_sn, port=kwargs.get('topas_port', 8000))
            except Exception as excpt:
                print("Exception reason: ", excpt)
                raise

            if self.tctrl is not None and self.tctrl.connected:
                print("Topas4 control established at {:}".format(
                    self.tctrl.url))
            else:
                print("Could not connect to the Topas4 REST server. An "
                      "incrrect port was likely specified. The default port "
                      "for the first Topas4 server is 8000, but in some cases"
                      "the server might be running at a higher port "
                      "(8000-8010). The program will continue, but some "
                      "advanced control features will be unavailable.")
                raise RuntimeError("Could not connect to Topas4 server")

        self.with_carbide_control = kwargs.get('with_carbide_control')
        if self.with_carbide_control:
            carbide_ip_address = kwargs.get('carbide_ip_address', '10.1.251.1')
            try:
                print("Connecting to the CARBIDE REST server...")
                self.cbxctrl = Carbide(
                    ip_address=carbide_ip_address, port=20010,
                    with_register_access=True)
            except Exception as excpt:
                print("Could not connected to the CARBIDE REST server")

            if self.cbxctrl is not None:
                print("CARBIDE control established at {:}".format(
                    self.cbxctrl.url))


    def __del__(self):
        self.logger.info("Stopping remote CRONUS-3P control")

    # === Status ===

    def get_status(self):
        """Get laser system status."""
        return self._get('Main/Status')

    def get_pump_laser_status(self):
        """Get pump laser satus."""
        return self._get('PumpLaser/Status')

    def get_info(self):
        """Get laser info."""
        return self._get('Main/Info')

    def wavelength_setting_state_str_to_enum(self, state_str):
        """Convert wavelength setting state string to enum."""
        c3p_state_strings = [
            'UnknownState', 'Fail', 'SoftFail',
            'Idle', 'Preparation', 'MovingMotorsToInitialPosition',
            'AdjustingOutputBeam', 'Success']

        if state_str not in c3p_state_strings:
            print("Unrecognized state string '{:s}'".format(state_str))
            return C3PState.UnknownState
        else:
            return C3PState(c3p_state_strings.index(state_str) - 3)

    def get_wavelength_setting_state(self):
        """Get wavelength setting state."""
        return self.wavelength_setting_state_str_to_enum(
            self.get_status()['WavelengthSettingState'])

    def print_status(self):
        """Print status."""
        status = self.get_status()

        print('Wavelength: {:.0f} nm'.format(status['Wavelength']))
        print('GDD: {:.0f} fs^2'.format(status['GDD']))

        print('Wavelength set result: ' + status['WavelengthSettingState'])

        print('Beam tracking:')
        beam_tr = status['BeamPositions'][0]
        print('\tNear: power={:.3f}, x={:.3f}, y={:.3f}'.format(
            beam_tr['Intensity'], beam_tr['X'], beam_tr['Y']))
        beam_tr = status['BeamPositions'][1]
        print('\tFar : power={:.3f}, x={:.3f}, y={:.3f}'.format(
            beam_tr['Intensity'], beam_tr['X'], beam_tr['Y']))

    # === Wavelength ===

    def get_wavelength(self):
        """Get wavelength in nm."""
        return self._get('Main/Status')['Wavelength']

    def set_wavelength(self, wavl, retry=True, num_retries=3, **kwargs):
        """Set wavelength in nm."""
        for attempt_ind in range(num_retries):
            try:
                return self._set_wavelength(wavl, **kwargs)
            except Exception as excpt:
                print("Could not set wavelength. Reason: ", excpt)
                if retry:
                    print("Trying again, attempt {:d}/{:d}".format(
                        attempt_ind, num_retries))
                if attempt_ind == num_retries:
                    print("Could not set wavelength after multiple retries")
                    raise excpt

    def _set_wavelength(
            self, wavl, gdd=None, skip_if_set=True, wait_until_done=False,
            with_beam_tracking=False, open_shutters=False, verbose=None):
        """Set wavelength with optional beam tracking."""
        if verbose is None:
            verbose = self.verbose

        if open_shutters:
            if verbose:
                print('Opening shutters...')
            self.open_all_shutters()

        old_wavl = self.get_wavelength()

        if self.has_gdd_control and gdd is not None:
            old_gdd = self.get_gdd()

        if skip_if_set and np.abs((old_wavl - wavl)) < 1:
            if not self.has_gdd_control or gdd is None:
                print("Wavelength already set")
                return
            elif np.abs((old_gdd - gdd)) < 100:
                print("Wavelength and GDD already set")
                return

        if verbose:
            if self.has_gdd_control and gdd is not None:
                print("Setting wavelength to {:.0f} nm and GDD "
                      "to {:.0f} fs2...".format(wavl, gdd),
                      end='', flush=True)
            else:
                print("Setting wavelength to {:.0f} nm...".format(wavl),
                      end='', flush=True)

        timestamp = datetime.datetime.now()
        timestamp_str = timestamp.strftime('%Y-%m-%d_%H%M%S')

        gdd_setting_enabled = False
        skip_gdd_setting_outside_range = True
        if self.has_gdd_control and gdd is not None:
            if wavl < 1250 or wavl > 1800:
                gdd_setting_enabled = False
            else:
                gdd_range = self.get_gdd_range()
                if gdd >= gdd_range[0] and gdd <= gdd_range[1]:
                    gdd_setting_enabled = True
                elif skip_gdd_setting_outside_range:
                    gdd_setting_enabled = True
                    print("Requested GDD is outside available range, setting "
                          "wavelength only")
                else:
                    gdd_setting_enabled = False

        if gdd_setting_enabled:
            self._put('Main/WavelengthAndGDD', json.dumps(
                {'Wavelength': float(wavl), 'GDD': float(gdd)}))
        else:
            self._put('Main/Wavelength', json.dumps(
                {'Wavelength': float(wavl)}))

        if wait_until_done:
            time.sleep(0.05)
            set_state = self.get_wavelength_setting_state()
            while set_state not in [C3PState.Success, C3PState.Fail]:
                if False:
                    print("Set state: " + set_state)
                time.sleep(0.05)
                set_state = self.get_wavelength_setting_state()
            if verbose:
                if set_state == C3PState.Success and \
                        self.get_wavelength() == wavl:
                    print('OK')
                elif set_state == C3PState.Fail:
                    raise RuntimeError("Failed to set wavelength with 'Fail' "
                                       "state")
                else:
                    print('Unspecified problem occurred')
        elif verbose:
            print('\n')

    def wait_until_wavelength_set(self):
        """Wait until wavelength setting procedure is completed."""
        while self.get_wavelength_setting_state() != 'Success':
            time.sleep(0.05)

    # === GDD ===

    def get_gdd(self):
        """Get GDD in fs^2."""
        # return self._get('GDD')['GDD']
        return self._get('Main/Status').get('GDD')

    def set_gdd(
            self, gdd, skip_if_set=True, wait_until_done=False,
            open_shutters=False, verbose=None):
        """Set GDD."""
        if verbose is None:
            verbose = self.verbose

        if not self.has_gdd_control:
            return False

        if open_shutters:
            if verbose:
                print('Opening shutters...')
            self.open_all_shutters()

        old_gdd = self.get_gdd()

        if skip_if_set and np.abs((old_gdd - gdd)) < 10:
            print("GDD already set")
            return True

        if verbose:
            print("Setting GDD to {:.0f} fs^2...".format(gdd), end='',
                  flush=True)

        timestamp = datetime.datetime.now()
        timestamp_str = timestamp.strftime('%Y-%m-%d_%H%M%S')

        # self._put('GDD', json.dumps({'GDD': float(gdd)}))
        self._put('Main/GDD', json.dumps({'GDD': float(gdd)}))

        if wait_until_done:
            time.sleep(0.05)
            set_state = self.get_wavelength_setting_state()
            while set_state not in [C3PState.Success, C3PState.Fail]:
                if False:
                    print("Set state: " + set_state)
                time.sleep(0.05)
                set_state = self.get_wavelength_setting_state()
            if verbose:
                if set_state == C3PState.Success and self.get_gdd() == gdd:
                    print('OK')
                elif set_state == C3PState.Fail:
                    # Hacky retry
                    print("Failed to set GDD, retrying with a wait...")
                    self._put('Main/GDD', json.dumps({'GDD': float(gdd)}))
                    time.sleep(3)
                    set_state = self.get_wavelength_setting_state()
                    if set_state == C3PState.Success:
                        print("OK")
                        return True
                    print("Failed")
                    return False
                else:
                    print('Unspecified problem occurred')
                    return False
        elif verbose:
            print('\n')
        return True

    def get_gdd_range(self, wavl=None):
        """Get GDD range in fs^2."""
        if wavl is None:
            response = self._get('Main/CurrentGDDRange')
            return [response.get('Min'), response.get('Max')]
        else:
            response = self._post('Main/GDDRange', {'Interaction': 'IDL',
                                                    'Wavelength': float(wavl)})
            return [response.get('Min'), response.get('Max')]

    # === Beam steering ===

    def has_beam_steering(self):
        """Check whether laser system has beam steering."""
        return self.get_info().get('HasOutputBeamStabilization', False)

    def is_beam_steering_active(self):
        """Check whether beam steering is active."""
        return self.get_status()['IsBeamSteeringActive']

    def get_beam_steering_pos(self):
        """Get the position of beam steering motors."""
        return self.get_status().get('BeamMirrorActualPositions')

    def get_beam_position(self):
        """Get beam position."""
        return self.get_status().get('BeamPositions')

    def get_beam_position_total(self):
        """Get beam position total signal."""
        status = self.get_status()
        qd1_ampl = status['BeamPositions'][0]['Intensity']
        qd2_ampl = status['BeamPositions'][1]['Intensity']

        return [qd1_ampl, qd2_ampl]

    def set_beam_pos(self, mirror_id=1, xpos=0, ypos=0, wait_until_done=False,
                     extra_wait_time=0.05):
        """Move beam steering mirror."""
        self._put('Beam/MoveBeam/{:d}'.format(mirror_id), json.dumps(
                  {'X': float(xpos), 'Y': float(ypos)}))

        if wait_until_done:
            print("Waiting for mirror to move...", end='', flush=True)
            t_start = time.time()
            while True:
                if time.time() - t_start > 5:
                    print("Problem\nMove taking longer than 5 s, reissuing "
                          "move command...")
                    self._put('Beam/MoveBeam/{:d}'.format(mirror_id),
                              json.dumps({'X': float(xpos), 'Y': float(ypos)}))
                    t_start = time.time()

                pos = self.get_beam_steering_pos()[mirror_id-1]
                if np.abs((pos['X'] - xpos)) < 0.01 \
                        and np.abs((pos['Y'] - ypos)) < 0.01:
                    break
                time.sleep(extra_wait_time)
            print('OK')

    def center_beam(self, wait_until_done=True):
        """Activate beam steering procedure to center the beam."""
        if wait_until_done:
            self.logger.info("Initiating beam steering...")
        else:
            self.logger.info("Beam steering initiated (no wait)")

        self._put('CenterBeam', '')

        if wait_until_done:
            while not self.is_beam_steering_active():
                time.sleep(0.01)

            while self.is_beam_steering_active():
                time.sleep(0.01)

            logger.info("Beam steering completed")


class C3Pv1Control(C3PControl_Base):
    def __init__(self, ip_address, **kwargs):
        self.hw_ver = 'c3pv1'
        super().__init__(ip_address, **kwargs)

    def print_status(self):
        """Print status for C3Pv1 systems."""
        super().print_status()
        status = self.get_status()

        print('Shutter: ' + 'Open' if status['IsPrimaryShutterOpen'] else
            'False')

    def is_output_shutter_open(self):
        return self.get_status()['IsPrimaryShutterOpen']

    def open_all_shutters(self, wait_to_stabilize=True, verbose=None,
                          max_wait=3):
        """Open all shutters."""
        if verbose is None:
            verbose = self.verbose

        if verbose:
            print("Opening shutters...")

        ret_val = self._put('Main/OpenShutters', '')
        t_start = time.time()

        time.sleep(1)

        while not self.get_status()['Shutters']['IsShutterOpen']:
            if time.time() - t_start > max_wait:
                raise RuntimeError("Could not open shutters in the allocated "
                                   "maximum time ({:d} seconds)".format(
                                       max_wait))
            time.sleep(0.1)

        if wait_to_stabilize:
            if verbose:
                print("Waiting for system to stabilize")
            time.sleep(2)

        return ret_val

    def close_all_shutters(self, wait_to_stabilize=False, verbose=None):
        """Close all shutters."""
        if verbose is None:
            verbose = self.verbose

        if verbose:
            print("Closing shutters...")

        ret_val = self._put('Main/CloseShutters', '')

        if wait_to_stabilize:
            if verbose:
                print("Waiting for system to stabilize")
            time.sleep(2)

        return ret_val

class C3Pv2Control(C3PControl_Base):
    def __init__(self, ip_address, dev_sn=None, **kwargs):
        self.hw_ver = 'c3pv2'
        super().__init__(ip_address, dev_sn=dev_sn, **kwargs)

        # Determine analog input adu to Volts factor
        try:
            file_name = "C:\Programs\WinTopas4\Resources\SelfHostedServer" + \
                    f"\{dev_sn}\Configuration\Settings\LightComposer" + \
                    "\Channels\IDL\ExternalAttenuatorControl.json"
            self.ain_adu_per_volt = json.load(open(file_name, 'r'))['InputAduPerVolt']
        except Exception as excpt:
            # TODO: This will fail for Light Composer <v3 because there is no 'Channels'
            # folder there yet. Might be good to scan the entire server
            # configuration folder to look for the file, or just use the
            # default.
            self.ain_adu_per_volt = 2625
            self.logger.warning("Could not determine adu to Volts from " \
                                "configuration, using default value " \
                                f"{self.ain_adu_per_volt:.1f}.")
            # raise

    def print_status(self):
        """Print status."""
        super().print_status()
        status = self.get_status()

        for shutter in status['Shutters']['Shutters']:
            print(f"Shuter '{shutter['Name']}' is {shutter['IsShutterOpen']}")

    def check_if_beam_is_ready(self):
        """Check if laser is ready for beam output."""
        # Make sure laser is ready
        status = self.get_status()
        if not status['IsPumpLaserReady']:
            raise RuntimeError("CRONUS-3P pump laser is not ready")
        if not status['IsWavelengthSet']:
            raise RuntimeError("CRONUS-3P wavelength is not set")

        return True

    # === Attenuator ===
    def get_attenuator_pos(self):
        """Get IDL channel attenuator motor position in steps."""
        if not self.check_attenuator():
            self.logger.error("Cannot get attenuator motor position.")
        else:
            return self.tctrl.get_motor_pos(self.idl_attn_motor_id)

    def set_attenuator_pos(self, pos, wait_until_done=False):
        """Set IDL channel attenuator motor position in steps."""
        if not self.check_attenuator():
            self.logger.error("Cannot get attenuator motor position.")
        else:
            self.tctrl.set_motor_pos(self.idl_attn_motor_id, pos)

            if wait_until_done:
                while True:
                    actual = self.tctrl.get_motor_pos(self.idl_attn_motor_id)
                    if np.abs((pos - actual)) < 10:
                        break
                    time.sleep(0.05)

    def check_attenuator(self):
        """Check whether IDL channel attenuator can be controlled."""
        if not self.tctrl:
            self.logger.error("Attenuator control requires Topas4 REST")
            return False

        if not self.idl_attn_motor_id:
            self.logger.error("IDL attenuator motor ID not defined")
            return False

        if not self.get_status().get('IsPumpLaserReady'):
            self.logger.error("Pump laser is not started")
            return False

        return True

    # === Shutters ===

    def is_output_shutter_open(self):
        """Check if the output shutter of the main tunable channel is open."""
        for shutter in self.get_status()['Shutters']['Shutters']:
            if shutter['Name'] == 'OPA':
                return shutter['IsShutterOpen']

        self.logger.error("Could not find 'OPA' output shutter")
        return False


    def open_all_shutters(self, wait_to_stabilize=True, verbose=None,
                          max_wait=3):
        """Open all shutters."""
        if verbose is None:
            verbose = self.verbose

        if verbose:
            print("Opening shutters...")

        ret_val = self._put('Main/OpenShutters', '')
        t_start = time.time()

        time.sleep(1)

        while not self.get_status()['Shutters']['IsShutterOpen']:
            if time.time() - t_start > max_wait:
                raise RuntimeError("Could not open shutters in the allocated "
                                   "maximum time ({:d} seconds)".format(
                                       max_wait))
            time.sleep(0.1)

        if wait_to_stabilize:
            if verbose:
                print("Waiting for system to stabilize")
            time.sleep(2)

        return ret_val

    def close_all_shutters(self, wait_to_stabilize=False, verbose=None):
        """Close all shutters."""
        if verbose is None:
            verbose = self.verbose

        if verbose:
            print("Closing shutters...")

        ret_val = self._put('Main/CloseShutters', '')

        if wait_to_stabilize:
            if verbose:
                print("Waiting for system to stabilize")
            time.sleep(2)

        return ret_val

    # === Analog input ===

    def _get_ain_adu(self):
        if not self.with_carbide_control:
            raise RuntimeError("AIN read requires CARBIDE control")
        if not self.with_topas_control:
            raise RuntimeError("AIN read requires Topas control")
        return self.cbxctrl.get_register('0x03000700')

    def get_ain_volts(self):
        return self._get_ain_adu()/self.ain_adu_per_volt
